/*
 * poke.c: the poker / accessible tree view
 *
 * Author:
 *    Michael Meeks
 *
 * Copyright 2002 Sun Microsystems, Inc.
 */

/*
 * TODO:
 *    + relation sets
 *    + whack a vertical divider into the app.
 *    + add 'getAccessibleAtPoint' on
 *      toplevel whatnot - if possible
 *    + Table view
 *    + Text selection view
 *    + AccessibleHypertext
 *    + AccessibleHyperlink
 *    + editable text
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <cspi/spi.h>
#include <glade/glade-xml.h>
#include <libgnomeui/gnome-ui-init.h>

#include "graphics.h"
#include "child-listener.h"
#include "accessible-tree-model.h"

extern void poke (Accessible *accessible);

typedef struct {
	GladeXML     *xml;

	Accessible   *root_node;

	GtkWidget    *window;

	GtkWidget    *accessible_role;
	GtkWidget    *accessible_role_image;
	GtkWidget    *accessible_relations_frame;
	GtkWidget    *accessible_states_frame;

	GtkTreeModel *tree_model;
	GtkTreeView  *tree_view;

	GtkListStore *state_set_store;
	GtkListStore *action_store;
	GtkWidget    *action_view;
	GtkWidget    *action_take;

	GtkWidget    *text;
	GtkWidget    *text_carret;
	gulong        text_changed_id;

	GtkWidget    *value_minimum;
	GtkWidget    *value_current;
	GtkWidget    *value_maximum;
	gulong        value_changed_id;

	GtkWidget    *image_descr;
	GtkWidget    *image_size;

	GtkListStore *selection_store;
	GtkTreeView  *selection_view;
	GtkWidget    *selection_all;
	GtkWidget    *selection_clear;
	gulong        selection_changed_id;

	Accessible   *selected;

	AccessibleCoordType ctype;
} Poker;

enum {
	POKER_SELECTION_ICON,
	POKER_SELECTION_NAME,
	POKER_SELECTION_DESCR,
	POKER_SELECTION_IDX
};

static void
update_relation_set (Poker *poker, AccessibleRelation **relations)
{
	int i;

	if (relations == NULL || *relations == NULL) {
		gtk_widget_hide (poker->accessible_relations_frame);
		return;
	}

	gtk_widget_show (poker->accessible_relations_frame);

	for (i = 0; relations [i]; i++) {
		AccessibleRelation *relation = relations [i];

		AccessibleRelation_unref (relation);
	}
	g_warning ("%d relations", i);
}

static void
update_if_accessible (Poker *poker, Accessible *accessible)
{
	int i;
	char *txt, *txt2;
	gboolean has_state;
	AccessibleRelation **relations;
	AccessibleStateSet *state_set;
	static const struct {
		AccessibleState state;
		const char     *name;
	} state_names [] = {
		{ SPI_STATE_INVALID, "INVALID" },
		{ SPI_STATE_ACTIVE, "ACTIVE" },
		{ SPI_STATE_ARMED, "ARMED" },
		{ SPI_STATE_BUSY, "BUSY" },
		{ SPI_STATE_CHECKED, "CHECKED" },
		{ SPI_STATE_COLLAPSED, "COLLAPSED" },
		{ SPI_STATE_DEFUNCT, "DEFUNCT" },
		{ SPI_STATE_EDITABLE, "EDITABLE" },
		{ SPI_STATE_ENABLED, "ENABLED" },
		{ SPI_STATE_EXPANDABLE, "EXPANDABLE" },
		{ SPI_STATE_EXPANDED, "EXPANDED" },
		{ SPI_STATE_FOCUSABLE, "FOCUSABLE" },
		{ SPI_STATE_FOCUSED, "FOCUSED" },
		{ SPI_STATE_HORIZONTAL, "HORIZONTAL" },
		{ SPI_STATE_ICONIFIED, "ICONIFIED" },
		{ SPI_STATE_MODAL, "MODAL" },
		{ SPI_STATE_MULTI_LINE, "MULTI_LINE" },
		{ SPI_STATE_MULTISELECTABLE, "MULTISELECTABLE" },
		{ SPI_STATE_OPAQUE, "OPAQUE" },
		{ SPI_STATE_PRESSED, "PRESSED" },
		{ SPI_STATE_RESIZABLE, "RESIZABLE" },
		{ SPI_STATE_SELECTABLE, "SELECTABLE" },
		{ SPI_STATE_SELECTED, "SELECTED" },
		{ SPI_STATE_SENSITIVE, "SENSITIVE" },
		{ SPI_STATE_SHOWING, "SHOWING" },
		{ SPI_STATE_SINGLE_LINE, "SINGLE_LINE" },
		{ SPI_STATE_STALE, "STALE" },
		{ SPI_STATE_TRANSIENT, "TRANSIENT" },
		{ SPI_STATE_VERTICAL, "VERTICAL" },
		{ SPI_STATE_VISIBLE, "VISIBLE" },
		{ 0, NULL }
	};

	txt = Accessible_getRoleName (accessible);
	txt2 = g_strdup_printf ("<big><b>%s</b></big>",
				txt && txt [0] != '\0' ? txt : "<no role>");
	gtk_label_set_markup (GTK_LABEL (poker->accessible_role), txt2);
	g_free (txt2);
	SPI_freeString (txt);
	
	gtk_image_set_from_pixbuf (
		GTK_IMAGE (poker->accessible_role_image),
		get_pixbuf_for_role (Accessible_getRole (accessible)));

	relations = Accessible_getRelationSet (accessible);
	update_relation_set (poker, relations);
	free (relations); /* FIXME: need a method for this */

	has_state = FALSE;
	gtk_list_store_clear (poker->state_set_store);
	state_set = Accessible_getStateSet (accessible);
	for (i = 0; state_names [i].name; i++) {
		if (AccessibleStateSet_contains (state_set, state_names [i].state)) {
			GtkTreeIter iter;
			g_warning ("Store append '%s'", state_names [i].name);
			gtk_list_store_append (poker->state_set_store, &iter);
			gtk_list_store_set (poker->state_set_store, &iter,
					    0, state_names [i].name, -1);
			has_state = TRUE;
		}
	}
	AccessibleStateSet_unref (state_set);

	if (has_state)
		gtk_widget_show (poker->accessible_states_frame);
	else
		gtk_widget_hide (poker->accessible_states_frame);
}

static void
component_grab_focus_clicked (GtkButton *button, Poker *poker)
{
	AccessibleComponent *component;

	g_return_if_fail (poker->selected != NULL);

	component = Accessible_getComponent (poker->selected);
	g_return_if_fail (component != NULL);

	g_warning ("Grab focus !");
	AccessibleComponent_grabFocus (component);

	AccessibleComponent_unref (component);
}

static void
update_if_component (Poker *poker, AccessibleComponent *component)
{
	int i;
	short zorder;
	long int x, y, w, h;
	char *txt;
	GtkWidget *widget;
	AccessibleComponentLayer layer;
	static const struct {
		AccessibleComponentLayer layer;
		const char              *name;
	} layer_names [] = {
		{ SPI_LAYER_INVALID, "INVALID" },
		{ SPI_LAYER_BACKGROUND, "BACKGROUND" },
		{ SPI_LAYER_CANVAS, "CANVAS" },
		{ SPI_LAYER_WIDGET, "WIDGET" },
		{ SPI_LAYER_MDI, "MDI" },
		{ SPI_LAYER_POPUP, "POPUP" },
		{ SPI_LAYER_OVERLAY, "OVERLAY" },
		{ 0, NULL }
	};

	AccessibleComponent_getExtents (
		component, &x, &y, &w, &h, poker->ctype);

	widget = glade_xml_get_widget (poker->xml, "component_if_position");
	txt = g_strdup_printf ("%ld, %ld", x, y);
	gtk_label_set_text (GTK_LABEL (widget), txt);
	g_free (txt);

	widget = glade_xml_get_widget (poker->xml, "component_if_size");
	txt = g_strdup_printf ("%ld, %ld", w, h);
	gtk_label_set_text (GTK_LABEL (widget), txt);
	g_free (txt);

	txt = "<invalid layer>";
	layer = AccessibleComponent_getLayer (component);
	for (i = 0; layer_names [i].name; i++) {
		if (layer_names [i].layer == layer)
			txt = (char *) layer_names [i].name;
	}
	widget = glade_xml_get_widget (poker->xml, "component_if_layer");
	gtk_label_set_text (GTK_LABEL (widget), txt);

	zorder = AccessibleComponent_getMDIZOrder (component);
	widget = glade_xml_get_widget (poker->xml, "component_if_zorder");
	txt = g_strdup_printf ("%d", (int) zorder);
	gtk_label_set_text (GTK_LABEL (widget), txt);
	g_free (txt);
}

static void
action_take_clicked (GtkButton *button, Poker *poker)
{
	AccessibleAction *action;
	GtkTreeSelection *selection;
	GtkTreeIter iter;

	g_return_if_fail (poker->selected != NULL);

	action = Accessible_getAction (poker->selected);
	g_return_if_fail (action != NULL);

	selection = gtk_tree_view_get_selection (
		GTK_TREE_VIEW (poker->action_view));
	if (selection
	    && gtk_tree_selection_get_selected (selection, NULL, &iter)) {
		GtkTreePath *path;
		path = gtk_tree_model_get_path (
			GTK_TREE_MODEL (poker->action_store), &iter);
		
		AccessibleAction_doAction (action, 
					   gtk_tree_path_get_indices (path)[0]);
		
		gtk_tree_path_free (path);
	}

	AccessibleAction_unref (action);
}

static void
action_row_activated_cb (GtkTreeView       *tree_view,
			 GtkTreePath       *path,
			 GtkTreeViewColumn *column,
			 Poker             *poker)
{
	action_take_clicked (NULL, poker);
}

static void
update_if_action (Poker *poker, AccessibleAction *action)
{
	long i, num_actions;
	GtkTreeIter iter;

	num_actions = AccessibleAction_getNActions (action);

	gtk_list_store_clear (poker->action_store);
	for (i = 0; i < num_actions; i++) {
		char *name, *descr, *keybinding;
		
		name = AccessibleAction_getName (action, i);
		descr = AccessibleAction_getDescription (action, i);
		keybinding = AccessibleAction_getKeyBinding (action, i);

		gtk_list_store_append (poker->action_store, &iter);
		gtk_list_store_set (poker->action_store, &iter,
				    0, name, 1, descr, 2, keybinding, -1);

		SPI_freeString (keybinding);
		SPI_freeString (descr);
		SPI_freeString (name);
	}

	if (gtk_tree_model_get_iter_root (GTK_TREE_MODEL (poker->action_store), &iter)) {
		gtk_tree_selection_select_iter (
			gtk_tree_view_get_selection (
				GTK_TREE_VIEW (poker->action_view)), &iter);
	}

	gtk_widget_set_sensitive (poker->action_take, num_actions > 0);
}

static void
update_if_text (Poker *poker, AccessibleText *text)
{
	char *txt;

	g_signal_handler_block (poker->text, poker->text_changed_id);

	txt = AccessibleText_getText (text, 0, -1);
	gtk_entry_set_text (GTK_ENTRY (poker->text), txt);
	SPI_freeString (txt);

	gtk_entry_set_editable (GTK_ENTRY (poker->text), FALSE);
	gtk_widget_set_sensitive (poker->text, FALSE);

	g_signal_handler_unblock (poker->text, poker->text_changed_id);

	gtk_spin_button_set_value (
		GTK_SPIN_BUTTON (poker->text_carret),
		AccessibleText_getCaretOffset (text));
	gtk_spin_button_set_range (
		GTK_SPIN_BUTTON (poker->text_carret),
		0, AccessibleText_getCharacterCount (text));
}

static void
editable_text_changed_cb (GtkEditable *editable_widget,
			  Poker       *poker)
{
	const char *txt;
	AccessibleEditableText *editable;

	g_return_if_fail (poker->selected != NULL);

	editable = Accessible_getEditableText (poker->selected);
	g_return_if_fail (editable != NULL);

	txt = gtk_entry_get_text (GTK_ENTRY (editable_widget));

	AccessibleEditableText_setTextContents (editable, txt);

	AccessibleEditableText_unref (editable);
}

static void
text_carret_changed_cb (GtkSpinButton *spin_button,
			Poker         *poker)
{
	AccessibleText *text;

	g_return_if_fail (poker->selected != NULL);

	text = Accessible_getText (poker->selected);
	g_return_if_fail (text != NULL);

	AccessibleText_setCaretOffset (
		text,
		gtk_spin_button_get_value (spin_button));

	AccessibleText_unref (text);
}


static void
update_if_editable (Poker *poker, AccessibleEditableText *editable)
{
	gtk_entry_set_editable (GTK_ENTRY (poker->text), TRUE);
	gtk_widget_set_sensitive (poker->text, TRUE);
}

static void
get_selection_n_editable (Poker                   *poker,
			  AccessibleEditableText **etext,
			  long int                *start,
			  long int                *end)
{
	AccessibleText *text;

	*etext = NULL;

	g_return_if_fail (poker->selected != NULL);
	
	text = Accessible_getText (poker->selected);

	AccessibleText_getSelection (text, 0, start, end);

	*etext = Accessible_getEditableText (poker->selected);

	AccessibleText_unref (text);
}

static void
update_text (Poker *poker)
{
	AccessibleText *text = Accessible_getText (poker->selected);

	g_return_if_fail (text != NULL);

	update_if_text (poker, text);
	update_if_editable (poker, NULL);

	AccessibleText_unref (text);
}

static void
editable_cut_clicked_cb (GtkButton *button, Poker *poker)
{
	long int start, end;
	AccessibleEditableText *etext;

	get_selection_n_editable (poker, &etext, &start, &end);
	g_return_if_fail (etext != NULL);

	AccessibleEditableText_cutText (etext, start, end);

	update_text (poker);

	AccessibleEditableText_unref (etext);
}

static void
editable_copy_clicked_cb (GtkButton *button, Poker *poker)
{
	long int start, end;
	AccessibleEditableText *etext;

	get_selection_n_editable (poker, &etext, &start, &end);
	g_return_if_fail (etext != NULL);

	AccessibleEditableText_copyText (etext, start, end);

	update_text (poker);

	AccessibleEditableText_unref (etext);
}

static void
editable_delete_clicked_cb (GtkButton *button, Poker *poker)
{
	long int start, end;
	AccessibleEditableText *etext;

	get_selection_n_editable (poker, &etext, &start, &end);
	g_return_if_fail (etext != NULL);

	AccessibleEditableText_deleteText (etext, start, end);

	update_text (poker);

	AccessibleEditableText_unref (etext);
}

static void
editable_paste_clicked_cb (GtkButton *button, Poker *poker)
{
	long int start, end;
	AccessibleEditableText *etext;

	get_selection_n_editable (poker, &etext, &start, &end);
	g_return_if_fail (etext != NULL);

	AccessibleEditableText_pasteText (etext, start);

	update_text (poker);

	AccessibleEditableText_unref (etext);
}

static void
value_text_changed_cb (GtkEditable *txt_widget,
		       Poker       *poker)
{
	const char *txt;
	AccessibleValue *value;

	g_return_if_fail (poker->selected != NULL);

	value = Accessible_getValue (poker->selected);
	g_return_if_fail (value != NULL);

	txt = gtk_entry_get_text (GTK_ENTRY (txt_widget));

	AccessibleValue_setCurrentValue (value, atof (txt));

	AccessibleValue_unref (value);
}

static void
update_if_value (Poker *poker, AccessibleValue *value)
{
	double v;
	char  *txt;

	v = AccessibleValue_getMinimumValue (value);
	txt = g_strdup_printf ("%g", v);
	gtk_label_set_text (GTK_LABEL (poker->value_minimum), txt);
	g_free (txt);

	v = AccessibleValue_getMaximumValue (value);
	txt = g_strdup_printf ("%g", v);
	gtk_label_set_text (GTK_LABEL (poker->value_maximum), txt);
	g_free (txt);

	g_signal_handler_block (poker->value_current,
				poker->value_changed_id);

	v = AccessibleValue_getCurrentValue (value);
	txt = g_strdup_printf ("%g", v);
	gtk_entry_set_text (GTK_ENTRY (poker->value_current), txt);
	g_free (txt);
	
	g_signal_handler_unblock (poker->value_current,
				  poker->value_changed_id);
}

static void
update_if_image (Poker *poker, AccessibleImage *image)
{
	char *txt;
	long int w, h;

	txt = AccessibleImage_getImageDescription (image);
	gtk_label_set_text (GTK_LABEL (poker->image_descr),
			    txt && txt [0] != '\0' ? txt : "<no description>");
	SPI_freeString (txt);

	AccessibleImage_getImageSize (image, &w, &h);
	txt = g_strdup_printf ("%ld, %ld", w, h);
	gtk_label_set_text (GTK_LABEL (poker->image_size), txt);
	g_free (txt);
}

static void
update_if_selection (Poker *poker, AccessibleSelection *selection)
{
	long i;
	long children;
	Accessible *parent;
	GtkTreeSelection *view_sel;

	parent = Accessible_queryInterface (
		selection, "IDL:Accessibility/Accessible:1.0");
	g_return_if_fail (parent != NULL);

	children = Accessible_getChildCount (parent);

	gtk_list_store_clear (poker->selection_store);
	view_sel = gtk_tree_view_get_selection (poker->selection_view);

	g_signal_handler_block (view_sel, poker->selection_changed_id);

	gtk_tree_selection_unselect_all (view_sel);

	for (i = 0; i < children; i++) {
		char *name, *descr;
		Accessible    *child;
		GtkTreeIter    iter;
		AccessibleRole role;
		GValue         idx;
		
		child = Accessible_getChildAtIndex (parent, i);

		name = Accessible_getName (child);
		role = Accessible_getRole (child);
		descr = Accessible_getDescription (child);

		if (!name || !name [0]) {
			AccessibleText *text = Accessible_getText (child);
			SPI_freeString (name);
			name = AccessibleText_getText (text, 0, -1);
			AccessibleText_unref (text);
		}

		memset (&idx, 0, sizeof (GValue));
		g_value_set_int (g_value_init (&idx, G_TYPE_INT), i);
		
		gtk_list_store_append (poker->selection_store, &iter);
		gtk_list_store_set (poker->selection_store, &iter,
				    POKER_SELECTION_NAME, name,
				    POKER_SELECTION_DESCR, descr,
				    POKER_SELECTION_ICON,
				    get_pixbuf_for_role (role),
				    POKER_SELECTION_IDX, &idx, -1);

		if (AccessibleSelection_isChildSelected (selection, i))
			gtk_tree_selection_select_iter (view_sel, &iter);

		SPI_freeString (descr);
		SPI_freeString (name);
		Accessible_unref (child);
	}

	g_signal_handler_unblock (view_sel, poker->selection_changed_id);

	Accessible_unref (parent);
}

static void
select_accessible_cb (GtkTreeModel *model,
		      GtkTreePath  *path,
		      GtkTreeIter  *iter,
		      gpointer      data)
{
	GValue *idx;
	AccessibleSelection *selection = data;

	gtk_tree_model_get (model, iter, POKER_SELECTION_IDX, &idx, -1);

	g_warning ("Select child %d", g_value_get_int (idx));
	AccessibleSelection_selectChild (selection, g_value_get_int (idx));
}

static void
selection_changed_cb (GtkTreeSelection *tree_selection, Poker *poker)
{
	AccessibleSelection *selection;

	selection = Accessible_getSelection (poker->selected);
	g_return_if_fail (selection != NULL);

	gtk_tree_selection_selected_foreach (
		tree_selection, select_accessible_cb, selection);

/*	update_if_selection (poker, selection); */

	AccessibleSelection_unref (selection);
}

static void
selection_clear_clicked_cb (GtkButton *button, Poker *poker)
{
	AccessibleSelection *selection;

	g_return_if_fail (poker->selected != NULL);

	selection = Accessible_getSelection (poker->selected);
	g_return_if_fail (selection != NULL);

	AccessibleSelection_clearSelection (selection);

	update_if_selection (poker, selection);

	AccessibleSelection_unref (selection);
}

static void
selection_all_clicked_cb (GtkButton *button, Poker *poker)
{
	AccessibleSelection *selection;

	g_return_if_fail (poker->selected != NULL);

	selection = Accessible_getSelection (poker->selected);
	g_return_if_fail (selection != NULL);

	AccessibleSelection_selectAll (selection);

	update_if_selection (poker, selection);

	AccessibleSelection_unref (selection);
}


static const struct {
	const char *widget_name;
	const char *interface_name;
	void (*update_fn) (Poker *poker, Accessible *accessible);
} poker_widget_names[] = {
	{ "poker_accessible_frame",    "IDL:Accessibility/Accessible:1.0",   update_if_accessible },
	{ "poker_application_label",   "IDL:Accessibility/Application:1.0",  NULL },
	{ "poker_streamable_frame",    "IDL:Accessibility/Streamable:1.0",   NULL },
	{ "poker_component_frame",     "IDL:Accessibility/Component:1.0",    update_if_component },
	{ "poker_action_frame",        "IDL:Accessibility/Action:1.0",       update_if_action },
	{ "poker_text_frame",          "IDL:Accessibility/Text:1.0",         update_if_text },
	{ "poker_editable_text_frame", "IDL:Accessibility/EditableText:1.0", update_if_editable },
	{ "poker_value_frame",         "IDL:Accessibility/Value:1.0",        update_if_value },
	{ "poker_table_frame",         "IDL:Accessibility/Table:1.0",        NULL }, /* FIXME */
	{ "poker_image_frame",         "IDL:Accessibility/Image:1.0",        update_if_image },
	{ "poker_selection_frame",     "IDL:Accessibility/Selection:1.0",    update_if_selection },
	{ NULL, NULL, NULL }
};

static void
update_frame_visibility (Poker *poker, Accessible *accessible)
{
	int i;

	for (i = 0; poker_widget_names [i].interface_name; i++) {
		GtkWidget *frame;
		AccessibleUnknown *unknown;

		unknown = Accessible_queryInterface (
			accessible, poker_widget_names [i].interface_name);

		if (poker_widget_names [i].widget_name)
			frame = glade_xml_get_widget (
				poker->xml, poker_widget_names [i].widget_name);
		else
			frame = NULL;

		if (unknown) {
			if (poker_widget_names [i].update_fn)
				poker_widget_names [i].update_fn (poker, unknown);
			if (frame)
				gtk_widget_show (frame);
		} else {
			if (frame)
				gtk_widget_hide (frame);
		}

		Accessible_unref (unknown);
	}
}

static void
poker_tree_selection_changed (GtkTreeSelection *selection,
			      Poker            *poker)
{
	GtkTreeIter iter;
	Accessible *accessible;
	GtkTreeModel *model;
	
	if (!gtk_tree_selection_get_selected (selection, &model, &iter)) {
		return;
	}

	accessible = accessible_tree_get_from_iter (&iter);

	poker->selected = accessible;

	update_frame_visibility (poker, accessible);
}

static void
refresh_clicked_cb (GtkButton *button, Poker *poker)
{
	update_frame_visibility (poker, poker->selected);
}

static void
expand_clicked_cb (GtkButton *button, Poker *poker)
{
	GtkTreeIter   iter;
	GtkTreePath  *path;
	GtkTreeModel *model;

	g_return_if_fail (poker->selected != NULL);

	gtk_tree_selection_get_selected (
		gtk_tree_view_get_selection (poker->tree_view),
		&model, &iter);

	path = gtk_tree_model_get_path (model, &iter);

	gtk_tree_view_expand_row (poker->tree_view, path, TRUE);

	gtk_tree_path_free (path);
}

static void
add_pix_name_column (GtkTreeView *tree_view,
		     int          icon_col,
		     int          name_col)
{
	GtkCellRenderer   *cell;
	GtkTreeViewColumn *column;

	column = gtk_tree_view_column_new ();
	gtk_tree_view_column_set_reorderable (column, TRUE);
	gtk_tree_view_column_set_title (column, "Name");

	cell = gtk_cell_renderer_pixbuf_new ();
	gtk_tree_view_column_pack_start (column, cell, FALSE);
	gtk_tree_view_column_set_attributes (
		column, cell, "pixbuf", icon_col, NULL);
	gtk_tree_view_column_set_resizable (column, TRUE);

	cell = gtk_cell_renderer_text_new ();
	gtk_tree_view_column_pack_start (column, cell, TRUE);
	gtk_tree_view_column_set_attributes (
		column, cell, "text", name_col, NULL);
	gtk_tree_view_column_set_sort_column_id (column, name_col);
	gtk_tree_view_append_column (tree_view, column);
}

static void
init_accessible_tree (Poker *poker, Accessible *accessible)
{
	GtkTreeIter iter;

	poker->tree_model = accessible_tree_model_new (accessible);

	poker->tree_view = GTK_TREE_VIEW (
		glade_xml_get_widget (
			poker->xml, "poker_tree_view"));

	gtk_tree_view_set_model (poker->tree_view, poker->tree_model);

	add_pix_name_column (poker->tree_view,
			     ACCESSIBLE_TREE_MODEL_ICON_COL,
			     ACCESSIBLE_TREE_MODEL_NAME_COL);

	gtk_tree_view_insert_column_with_attributes
		(poker->tree_view,
		 ACCESSIBLE_TREE_MODEL_ROLE_COL, "Role",
		 gtk_cell_renderer_text_new (),
		 "text", 1,
		 NULL);

	gtk_tree_view_insert_column_with_attributes
		(poker->tree_view,
		 ACCESSIBLE_TREE_MODEL_DESCR_COL, "Description",
		 gtk_cell_renderer_text_new (),
		 "text", 2,
		 NULL);

	gtk_tree_view_columns_autosize (poker->tree_view);
	gtk_tree_selection_set_mode (
		gtk_tree_view_get_selection (
			poker->tree_view),
		GTK_SELECTION_BROWSE);
	g_signal_connect (
		gtk_tree_view_get_selection (poker->tree_view),
		"changed",
		G_CALLBACK (poker_tree_selection_changed),
		poker);

	if (gtk_tree_model_get_iter_root (
		GTK_TREE_MODEL (poker->tree_model), &iter)) {
		GtkTreePath *path;

		gtk_tree_selection_select_iter (
			gtk_tree_view_get_selection (poker->tree_view), &iter);

		path = gtk_tree_model_get_path (poker->tree_model, &iter);
		gtk_tree_view_expand_row (poker->tree_view, path, FALSE);
		gtk_tree_path_free (path);
	}
}

static void
root_died_cb (ChildListener *listener, Accessible *root, Poker *poker)
{
	if (root == poker->root_node)
		gtk_main_quit ();
}

void
poke (Accessible *accessible)
{
	Poker *poker;
	GtkWidget *widget;

	poker = g_new0 (Poker, 1);

	/* FIXME: need a toggle for this */
	poker->ctype = SPI_COORD_TYPE_WINDOW;

	poker->root_node = accessible;
	poker->xml = get_glade_xml ();

	{ /* accessible interface */
		poker->accessible_role = glade_xml_get_widget (
			poker->xml, "accessible_if_role");
		poker->accessible_role_image = glade_xml_get_widget (
			poker->xml, "accessible_if_role_image");
		poker->accessible_relations_frame = glade_xml_get_widget (
			poker->xml, "accessible_if_relations_frame");
		poker->accessible_states_frame = glade_xml_get_widget (
			poker->xml, "accessible_if_states_frame");

		poker->state_set_store = gtk_list_store_new (1, G_TYPE_STRING);
		widget = glade_xml_get_widget (poker->xml, "accessible_if_state_list");
		gtk_tree_view_set_model (GTK_TREE_VIEW (widget),
					 GTK_TREE_MODEL (poker->state_set_store));
		gtk_tree_view_insert_column_with_attributes (
			GTK_TREE_VIEW (widget), 0, "State name",
			gtk_cell_renderer_text_new (), "text", 0, NULL);
	}

	{ /* component interface */
		g_signal_connect (
			glade_xml_get_widget (poker->xml, "component_if_grab_focus"),
			"clicked",
			G_CALLBACK (component_grab_focus_clicked),
			poker);
	}

	{ /* action interface */
		poker->action_store = gtk_list_store_new (
			3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);

		poker->action_view = widget = glade_xml_get_widget (
			poker->xml, "action_if_list");

		gtk_tree_view_set_model (GTK_TREE_VIEW (widget),
					 GTK_TREE_MODEL (poker->action_store));
		gtk_tree_view_insert_column_with_attributes (
			GTK_TREE_VIEW (widget), 0, "Name",
			gtk_cell_renderer_text_new (), "text", 0, NULL);
		gtk_tree_view_insert_column_with_attributes (
			GTK_TREE_VIEW (widget), 1, "Descr",
			gtk_cell_renderer_text_new (), "text", 0, NULL);
		gtk_tree_view_insert_column_with_attributes (
			GTK_TREE_VIEW (widget), 2, "Keybinding",
			gtk_cell_renderer_text_new (), "text", 0, NULL);
		g_signal_connect (widget, "row_activated",
				  G_CALLBACK (action_row_activated_cb),
				  poker);

		poker->action_take = glade_xml_get_widget (
			poker->xml, "action_if_take");
		g_signal_connect (
			poker->action_take, "clicked",
			G_CALLBACK (action_take_clicked), poker);
	}

	{ /* text interface */
		poker->text = glade_xml_get_widget (
			poker->xml, "text_if_entry");
		poker->text_changed_id = 
			g_signal_connect (poker->text, "changed",
					  G_CALLBACK (editable_text_changed_cb),
					  poker);
		poker->text_carret = glade_xml_get_widget (
			poker->xml, "text_if_carret");
		g_signal_connect (poker->text_carret,
				  "value_changed",
				  G_CALLBACK (text_carret_changed_cb),
				  poker);
	}

	{ /* editable text interface */
		g_signal_connect (
			glade_xml_get_widget (
				poker->xml, "editable_if_cut"),
			"clicked",
			G_CALLBACK (editable_cut_clicked_cb),
			poker);
		g_signal_connect (
			glade_xml_get_widget (
				poker->xml, "editable_if_copy"),
			"clicked",
			G_CALLBACK (editable_copy_clicked_cb),
			poker);
		g_signal_connect (
			glade_xml_get_widget (
				poker->xml, "editable_if_paste"),
			"clicked",
			G_CALLBACK (editable_paste_clicked_cb),
			poker);
		g_signal_connect (
			glade_xml_get_widget (
				poker->xml, "editable_if_delete"),
			"clicked",
			G_CALLBACK (editable_delete_clicked_cb),
			poker);
	}

	{ /* value interface */
		poker->value_minimum = glade_xml_get_widget (
			poker->xml, "value_if_minimum");
		poker->value_current = glade_xml_get_widget (
			poker->xml, "value_if_current");
		poker->value_maximum = glade_xml_get_widget (
			poker->xml, "value_if_maximum");
		poker->value_changed_id = 
			g_signal_connect (poker->value_current,
					  "changed",
					  G_CALLBACK (value_text_changed_cb),
					  poker);
	}

	{ /* image interface */
		poker->image_descr = glade_xml_get_widget (
			poker->xml, "image_if_description");
		poker->image_size  = glade_xml_get_widget (
			poker->xml, "image_if_size");
	}

	{ /* selection interface */
		poker->selection_store = gtk_list_store_new (
			4, GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_VALUE);

		widget = glade_xml_get_widget (
			poker->xml, "selection_if_children");
		poker->selection_view = GTK_TREE_VIEW (widget); 

		gtk_tree_view_set_model (GTK_TREE_VIEW (widget),
					 GTK_TREE_MODEL (poker->selection_store));
		gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (widget), FALSE);
		add_pix_name_column (GTK_TREE_VIEW (widget),
				     POKER_SELECTION_ICON,
				     POKER_SELECTION_NAME);
		gtk_tree_view_insert_column_with_attributes (
			GTK_TREE_VIEW (widget), 2, "Descr",
			gtk_cell_renderer_text_new (), "text",
			POKER_SELECTION_DESCR, NULL);

		poker->selection_all = glade_xml_get_widget (
			poker->xml, "selection_if_all");
		poker->selection_clear = glade_xml_get_widget (
			poker->xml, "selection_if_clear");

		poker->selection_changed_id = g_signal_connect (
			gtk_tree_view_get_selection (poker->selection_view),
			"changed", G_CALLBACK (selection_changed_cb), poker);
		g_signal_connect (
			glade_xml_get_widget (poker->xml, "selection_if_clear"),
			"clicked", G_CALLBACK (selection_clear_clicked_cb), poker);
		g_signal_connect (
			glade_xml_get_widget (poker->xml, "selection_if_all"),
			"clicked", G_CALLBACK (selection_all_clicked_cb), poker);
	}

	{ /* main frame buttons */
		g_signal_connect (
			glade_xml_get_widget (
				poker->xml, "poker_refresh"),
			"clicked",
			G_CALLBACK (refresh_clicked_cb),
			poker);

		g_signal_connect (
			glade_xml_get_widget (
				poker->xml, "poker_expand"),
			"clicked",
			G_CALLBACK (expand_clicked_cb),
			poker);
	}

	{ /* main window */
		GClosure *closure;

		poker->window = glade_xml_get_widget (
			poker->xml, "poker_window");

		closure = g_cclosure_new (
			G_CALLBACK (root_died_cb), poker, NULL);
		g_object_watch_closure (
			G_OBJECT (poker->window), closure);
		g_signal_connect_closure (
			child_listener_get (), "root_died", closure, FALSE);

		child_listener_add_root (child_listener_get (), poker->root_node);
	}

	init_accessible_tree (poker, accessible);
	gtk_widget_show (poker->window);

	gtk_main ();

	child_listener_remove_root (child_listener_get (), poker->root_node);

	gtk_widget_destroy (poker->window);

	g_object_unref (poker->xml);
	g_free (poker);
}
